@@ -9,9 +9,9 @@ use Data::Dump::OneLine qw(dump1);
 use Moo;
 use Perinci::Object;
 use Perinci::ToUtil;
-use Scalar::Util qw(reftype);
+use Scalar::Util qw(reftype blessed);
 
-our $VERSION = '0.71'; # VERSION
+our $VERSION = '0.72'; # VERSION
 
 with 'Perinci::To::Text::AddDocLinesRole';
 with 'SHARYANTO::Role::Doc::Section';
@@ -55,8 +55,11 @@ has undo_dir => (
 );
 
 has format => (is => 'rw', default=>sub{'text'});
+# bool, is format set via cmdline opt?
+has format_set => (is => 'rw');
 has format_options => (is => 'rw');
-has format_options_set => (is => 'rw'); # bool
+# bool, is format_options set via cmdline opt?
+has format_options_set => (is => 'rw');
 has pa_args => (is => 'rw');
 has _pa => (
     is => 'rw',
@@ -105,30 +108,59 @@ sub BUILD {
     #$self->{indent} = $args->{indent} // "    ";
 }
 
-sub format_and_display_result {
+sub format_result {
     require Perinci::Result::Format;
 
     my $self = shift;
-    return unless $self->{_res};
+    my $res  = $self->{_res};
+    return unless $res;
 
-    my $resmeta = $self->{_res}[3] // {};
-    unless ($resmeta->{"cmdline.display_result"}//1) {
-        $self->{_res}[2] = undef;
+    my $resmeta = $res->[3] // {};
+    unless ($resmeta->{"cmdline.display_result"} // 1) {
+        $res->[2] = undef;
+        return;
     }
 
-    my $format = $self->format;
+    my $format = $self->format_set ?
+        $self->format :
+            $self->{_meta}{"x.perinci.cmdline.default_format"} // $self->format;
     die "ERROR: Unknown output format '$format', please choose one of: ".
         join(", ", sort keys(%Perinci::Result::Format::Formats))."\n"
             unless $Perinci::Result::Format::Formats{$format};
     if ($self->format_options_set) {
-        $self->{_res}[3]{result_format_options} = $self->format_options;
+        $resmeta->{result_format_options} = $self->format_options;
     }
-    $log->tracef("Formatting output with %s", $format);
 
-    $self->{_fres} = Perinci::Result::Format::format(
-        $self->{_res}, $format);
+    if ($resmeta->{is_stream}) {
+        $log->tracef("Result is a stream");
+    } else {
+        $log->tracef("Formatting output with %s", $format);
+        $self->{_fres} = Perinci::Result::Format::format(
+            $self->{_res}, $format);
+    }
+}
 
-    # display result
+sub display_result {
+    require File::Which;
+
+    my $self = shift;
+
+    my $res  = $self->{_res};
+    return unless $res;
+
+    my $resmeta = $res->[3] // {};
+
+    # this is a specific (and wrong) workaround to avoid the "Wide character in
+    # print" error when printing, even though something like binmode(STDOUT,
+    # ":utf8") has been . this only happens when using the Console (text-pretty)
+    # formatter. i haven't managed to pinpoint who the culprit is (it's not
+    # Term::Size and probably something that Text::ASCIITable uses). i will
+    # remove this workaround because i will be replacing Text::ASCIITable with
+    # Text::ANSITable sometime soon anyway.
+    local *STDOUT = \*STDOUT;
+    binmode(STDOUT, ":utf8");
+
+    my $handle;
     if ($resmeta->{"cmdline.page_result"}) {
         my $pager = $resmeta->{"cmdline.pager"} //
             $ENV{PAGER};
@@ -142,22 +174,35 @@ sub format_and_display_result {
             die "Can't determine PAGER";
         }
         $log->tracef("Paging output using %s", $pager);
-        open my($p), "| $pager";
-        print $p $self->{_fres};
+        open $handle, "| $pager";
     } else {
-        {
-            # this is a specific (and wrong) workaround to avoid the "Wide
-            # character in print" error when printing, even though something
-            # like binmode(STDOUT, ":utf8") has been . this only happens when
-            # using the Console (text-pretty) formatter. i haven't managed to
-            # pinpoint who the culprit is (it's not Term::Size and probably
-            # something that Text::ASCIITable uses). i will remove this
-            # workaround because i will be replacing Text::ASCIITable with
-            # Text::ANSITable sometime soon anyway.
-            local *STDOUT = \*STDOUT;
-            binmode(STDOUT, ":utf8");
-            print $self->{_fres};
+        $handle = \*STDOUT;
+    }
+
+    if ($resmeta->{is_stream}) {
+        die "Can't format stream as " . $self->format .
+            ", please use --format text\n" unless $self->format =~ /^text/;
+        my $r = $res->[2];
+        if (ref($r) eq 'GLOB') {
+            while (!eof($r)) {
+                print $handle ~~<$r>;
+            }
+        } elsif (blessed($r) && $r->can('getline') && $r->can('eof')) {
+            # IO::Handle-like object
+            while (!$r->eof) {
+                print $r->getline;
+            }
+        } elsif (ref($r) eq 'ARRAY') {
+            # tied array
+            while (~~(@$r) > 0) {
+                print shift(@$r);
+            }
+        } else {
+            die "Invalid stream in result (not a glob/IO::Handle-like object/".
+                "(tied) array)\n";
         }
+    } else {
+        print $handle $self->{_fres};
     }
 }
 
@@ -777,7 +822,10 @@ sub gen_common_opts {
             $self->{_check_required_args} = 0;
         },
 
-        "format=s" => sub { $self->format($_[1]) },
+        "format=s" => sub {
+            $self->format_set(1);
+            $self->format($_[1]);
+        },
         "format-options=s" => sub {
             $self->format_options_set(1);
             $self->format_options(__json_decode($_[1]));
@@ -1087,7 +1135,8 @@ sub run {
         $log->tracef("<- %s(), return=%s", $meth, $exit_code);
         last if defined $exit_code;
     }
-    $self->format_and_display_result;
+    $self->format_result;
+    $self->display_result;
 
     $log->tracef("<- CmdLine's run(), exit code=%s", $exit_code);
     if ($self->exit) {
@@ -1111,7 +1160,7 @@ Perinci::CmdLine - Rinci/Riap-based command-line application framework
 
 =head1 VERSION
 
-version 0.71
+version 0.72
 
 =head1 SYNOPSIS
 
@@ -1198,10 +1247,7 @@ This module uses L<Log::Any> and L<Log::Any::App> for logging.
 
 This module uses L<Moo> for OO.
 
-
-This module has L<Rinci> metadata.
-
-=for Pod::Coverage ^(BUILD|run_.+|doc_.+|before_.+|after_.+|format_and_display_result|gen_common_opts|get_subcommand|list_subcommands|parse_common_opts|parse_subcommand_opts|format_options|format_options_set)$
+=for Pod::Coverage ^(BUILD|run_.+|doc_.+|before_.+|after_.+|format_result|display_result|gen_common_opts|get_subcommand|list_subcommands|parse_common_opts|parse_subcommand_opts|format_set|format_options|format_options_set)$
 
 =head1 ATTRIBUTES
 
@@ -1435,10 +1481,54 @@ C<progress> feature to true), Perinci::CmdLine will setup an output, currently
 either L<Progress::Any::Output::TermProgressBar> if program runs interactively,
 or L<Progress::Any::Output::LogAny> if program doesn't run interactively.
 
+=head1 METADATA PROPERTY ATTRIBUTE
+
+This module observes the following Rinci metadata property attributes:
+
+=head2 x.perinci.cmdline.default_format => STR
+
+Set default output format (if user does not specify via --format command-line
+option).
+
 =head1 RESULT METADATA
 
 This module interprets the following result metadata keys:
 
+=head2 is_stream => BOOL
+
+XXX should perhaps be defined as standard in L<Rinci::function>.
+
+If set to 1, signify that result is a stream. Result must be a glob, or an
+object that responds to getline() and eof() (like a Perl L<IO::Handle> object),
+or an array/tied array. Format must currently be C<text> (streaming YAML might
+be supported in the future). Items of result will be displayed to output as soon
+as it is retrieved, and unlike non-streams, it can be infinite.
+
+An example function:
+
+ $SPEC{cat_file} = { ... };
+ sub cat_file {
+     my %args = @_;
+     open my($fh), "<", $args{path} or return [500, "Can't open file: $!"];
+     [200, "OK", $fh, {is_stream=>1}];
+ }
+
+another example:
+
+ use Tie::Simple;
+ $SPEC{uc_file} = { ... };
+ sub uc_file {
+     my %args = @_;
+     open my($fh), "<", $args{path} or return [500, "Can't open file: $!"];
+     my @ary;
+     tie @ary, "Tie::Simple", undef,
+         SHIFT     => sub { eof($fh) ? undef : uc(~~<$fh> // "") },
+         FETCHSIZE => sub { eof($fh) ? 0 : 1 };
+     [200, "OK", \@ary, {is_stream=>1}];
+ }
+
+See also L<Data::Unixish> and L<App::dux> which deals with streams.
+
 =head2 cmdline.display_result => BOOL
 
 If you don't want to display function output (for example, function output is a
@@ -1534,13 +1624,20 @@ value will be retrieved from standard input if not specified. Example:
  };
  sub cmd {
      my %args = @_;
-     [200, "OK", "arg is $args{arg}"];
+     [200, "OK", "arg is '$args{arg}'"];
  }
  Perinci::CmdLine->new(url=>'/main/cmd')->run;
 
 When run from command line:
 
- % cmd --arg v1
+ % cat file.txt
+ This is content of file.txt
+ % cat file.txt | cmd
+ arg is 'This is content of file.txt'
+
+=head2 But I don't want it slurped into a single scalar, I want streaming!
+
+See L<App::dux> for an example on how to accomplish that.
 
 =head1 SEE ALSO
 
@@ -1562,8 +1659,5 @@ the same terms as the Perl 5 programming language system itself.
 
 =head1 FUNCTIONS
 
-
-None are exported by default, but they are exportable.
-
 =cut
 
